library(TDboost)
library(ktweedie)
library(MASS)
library(tweedie)
library(statmod)
library(mgcv)
library(beepr)
library(knitr)

########## n.sim is the number of simulation replications ##########
###### set it to some larger number for test error and timing ######
############# A plot is generated for each replication #############
############### Sound alarm at the end of the script ###############

### Edit if needed ###
n.sim <- 5 # Number of simulation
phi <- 0.5  # Tweedie dispersion parameter in (0, inf)
rho <- 1.5  # Tweedie power parameter in (1, 2)
N <- 400 # sample size: ntrain = ntest = N
sound_alarm_when_finish <- FALSE # play a sound when code execution finishes.
### Edit if needed ###





P <- 1  # dimension
# seed <- 20220625
# set.seed(seed)

# Pre-allocation for result storage
sim1 <- sim2 <- sim3 <- sim4 <- sim5 <- rep(NA, n.sim)
timing1 <- timing2 <- timing3 <- timing4 <- timing5 <- rep(NA, n.sim)

rand_tweedie<- function(mu,...) {
  Y <- rtweedie(1, mu = mu,...)
  Y
}

# Ranges from which candidate lam1 and sigma are chosen
lambda.range.rbf <- c(1e-30, 1e-10)
lambda.range.lap <- c(1e-3, 1e-2)

sigma.range.rbf <- c(1e-1, 1e1)
sigma.range.lap <- c(1e-1, 1e0)


for (i in 1:n.sim) {
  cat("Simulation no.", i, "\n")

  xtrain <- matrix(runif(N * P, 0, 1), N, P)
  xtest <- matrix(runif(N * P, 0, 1), N, P)
  Fxtrain <- 0.5 * (xtrain > 0.5)
  Fxtest <- 0.5 * (xtest > 0.5)
  truef <- as.vector(Fxtest)
  
  train_dat <- data.frame(xtrain)
  test_dat = data.frame(xtest)
  colnames(train_dat) <- colnames(test_dat) <- c("X1")
  
  mu <- exp(Fxtrain)
  ytrain <- sapply(mu, rand_tweedie, xi = rho, phi = phi)
  train_dat$Y <- ytrain
  
  # MGCV/TGAM
  t0 <- proc.time()
  m1 = gam(Y ~ s(X1), data = train_dat, family=Tweedie(p=rho, link = "log"))
  pred_f1 = as.vector(predict.gam(m1, newdata = test_dat))
  timing1[i] <- (proc.time() - t0)[3]

  # TDboost
  t0 <- proc.time()
  m2 <- TDboost(Y ~ X1, data = train_dat,
                distribution = list(name="EDM",alpha=rho), # poisson, coxph, and quantile available
                shrinkage = 0.001,               # shrinkage or learning rate,
                n.trees = 10000,              # 0.001 to 0.1 usually work
                interaction.depth = 1,         # 1: additive model, 2: two-way interactions, etc.
                bag.fraction = 1,            # subsampling fraction, 0.5 is probably best
                n.minobsinnode = 20,		 # minimum total weight needed in each node
                keep.data = TRUE,              # keep a copy of the dataset with the object
                cv.folds = 5,
                verbose = FALSE)                # print out progress
  best.iter <- TDboost.perf(m2, method="cv", plot.it = F)
  pred_f2 <- predict.TDboost(m2, test_dat, best.iter, type = "link")
  timing2[i] <- (proc.time() - t0)[3]
  
  # TGLM
  t0 <- proc.time()
  m3 <- glm(Y ~ X1, data = train_dat,
            family = tweedie(link.power=0, var.power=rho))
  pred_f3 <- predict(m3, test_dat)
  timing3[i] <- (proc.time() - t0)[3]

  ### ktweedie RBF kernel
  t0 <- proc.time()
  ## cross validation
  ktd.cv.rbf <- ktd_cv2d(x = xtrain,
                         y = ytrain,
                         kernfunc = rbfdot,
                         lambda = lambda.range.rbf,
                         sigma = sigma.range.rbf,
                         ncoefs = 15,
                         nfolds = 5,
                         ftol = 1e-4,
                         partol = 1e-4,
                         rho = rho,
                         verbose = FALSE)
  # training
  m4 <- ktd_estimate(x = xtrain,
                     y = ytrain,
                     kern = rbfdot(ktd.cv.rbf$Best_sigma),
                     lam1 = ktd.cv.rbf$Best_lambda,
                     rho = rho,
                     ftol = 1e-6,
                     partol = 1e-6,
                     verbose = FALSE)
  # predicting
  pred_m4 <- ktd_predict(model = m4, newdata = xtest, type = "link")
  pred_f4 <- as.vector(pred_m4$prediction)
  timing4[i] <- (proc.time() - t0)[3]
  
  ### ktweedie Laplace kernel
  t0 <- proc.time()
  ## cross validation
  ktd.cv.lap <- ktd_cv2d(x = xtrain,
                         y = ytrain,
                         kernfunc = laplacedot,
                         lambda = lambda.range.lap,
                         sigma = sigma.range.lap,
                         ncoefs = 15,
                         nfolds = 5,
                         ftol = 1e-4,
                         partol = 1e-4,
                         rho = rho,
                         verbose = FALSE)
  # training
  m5 <- ktd_estimate(x = xtrain,
                     y = ytrain,
                     kern = laplacedot(ktd.cv.lap$Best_sigma),
                     lam1 = ktd.cv.lap$Best_lambda,
                     rho = rho,
                     ftol = 1e-6,
                     partol = 1e-6,
                     verbose = FALSE)
  # predicting
  pred_m5 <- ktd_predict(model = m5, newdata = xtest, type = "link")
  pred_f5 <- as.vector(pred_m5$prediction)
  timing5[i] <- (proc.time() - t0)[3]

  # Mean absolute difference (MAD) storage
  sim1[i] <- mean(abs(pred_f1 - truef))
  sim2[i] <- mean(abs(pred_f2 - truef))
  sim3[i] <- mean(abs(pred_f3 - truef))
  sim4[i] <- mean(abs(pred_f4 - truef))
  sim5[i] <- mean(abs(pred_f5 - truef))

  # Plotting
  hx = 0.5 * (xtest > 0.5)
  par(mfrow = c(1,6), mar = c(4,4,4,1) + 0.1)
  limit = c(-0.1, 0.6)
  plot(y = hx, x = as.vector(xtest),
       xlab = expression(paste(x)), ylab = expression(paste(F(x))),
       ylim = limit,
       main = "(a) TRUE", pch = ".", cex = 2)
  plot(y = pred_f1, x = as.vector(xtest),
       xlab = expression(paste(x)), ylab = expression(paste(F(x))),
       ylim = limit,
       main = "(b) MGCV", pch = ".", cex = 2)
  plot(y = pred_f2, x = as.vector(xtest),
       xlab = expression(paste(x)), ylab = expression(paste(F(x))),
       ylim = limit,
       main = "(c) TDboost", pch = ".", cex = 2)
  plot(y = pred_f3, x = as.vector(xtest),
       xlab = expression(paste(x)), ylab = expression(paste(F(x))),
       ylim = limit,
       main = "(d) TGLM", pch = ".", cex = 2)
  plot(y = pred_f4, x = as.vector(xtest),
       xlab = expression(paste(x)), ylab = expression(paste(F(x))),
       ylim = limit,
       main = "(e) RBF", pch = ".", cex = 2)
  plot(y = pred_f5, x = as.vector(xtest),
       xlab = expression(paste(x)), ylab = expression(paste(F(x))),
       ylim = limit,
       main = "(f) Laplace", pch = ".", cex = 2)
}

# Sound alarm for completion
if (sound_alarm_when_finish) {
  beep(8)
}

# Compare results
MAD_mean <- c(mean(sim1), mean(sim2), mean(sim3), mean(sim4), mean(sim5))
MAD_se <- c(sd(sim1), sd(sim2), sd(sim3), sd(sim4), sd(sim5))/sqrt(n.sim - 1)

Time_mean <- c(mean(timing1),
               mean(timing2),
               mean(timing3),
               mean(timing4),
               mean(timing5))
Time_se <- c(sd(timing1),
             sd(timing2),
             sd(timing3),
             sd(timing4),
             sd(timing5))

models <- c("MGCV", "TDboost", "TGLM", "RBF", "Laplace")
names(MAD_mean) <- names(MAD_se) <- names(Time_mean) <- names(Time_se) <- models

# MAD results
kable(cbind(MAD_mean, MAD_se), format = "simple",
      caption = "Mean Absolute Difference")
# Timing results
kable(cbind(Time_mean, Time_se), format = "simple",
      caption = "Timing Results")









