library(TDboost)
library(ktweedie)
library(MASS)
library(tweedie)
library(statmod)
library(mgcv)
library(beepr)
library(lattice)
library(gridExtra)
library(knitr)


########## n.sim is the number of simulation replications ##########
########## set it to 1 and run the whole script for plots ##########
###### set it to some larger number for test error and timing ######
############### Sound alarm at the end of the script ###############

### Edit if needed ###
n.sim <- 5 # Number of simulation
phi <- 0.5  # Tweedie dispersion parameter
rho <- 1.5  # Tweedie power parameter
N <- 400 # sample size: ntrain = ntest = N
sound_alarm_when_finish <- FALSE # play a sound when code execution finishes.
### Edit if needed ###





P <- 2  # dimension

sim1 <- sim2 <- sim3 <- sim4 <- sim5 <- rep(NA, n.sim)
timing1 <- timing2 <- timing3 <- timing4 <- timing5 <- rep(NA, n.sim)

rand_tweedie<- function(mu,...) {
  Y <- rtweedie(1, mu = mu,...)
  Y
}

h2 <- function(x) {
  exp( -5 * (1-x[,1])^2 + (x[,2])^2 ) + exp( -5 * (x[,1])^2 + (1-x[,2])^2 )
}

lambda.range.rbf <- c(1e-30, 1e-10)
# sigma.range.rbf <- c(1,5)   # estimated using kernlab::sigest()
lambda.range.lap <- c(1e-3, 1e-2)
sigma.range.lap <- c(1, 2)

# plotting
plot_X1 = seq(0,1,0.03)
plot_X2 = seq(0,1,0.03)
plot_df <- expand.grid(plot_X1, plot_X2); names(plot_df) <- c("V1", "V2")
plot_mat = as.matrix(plot_df)

plot_true <- h2(plot_mat)

for (i in 1:n.sim) {
  xtrain <- matrix(runif(N * P, 0, 1), N, P)
  xtest <- matrix(runif(N * P, 0, 1), N, P)
  Fxtrain <- h2(xtrain)
  Fxtest <- h2(xtest)
  truef <- Fxtest
  
  train_dat <- data.frame(xtrain)
  test_dat = data.frame(xtest)
  colnames(train_dat) <- colnames(test_dat) <- c("V1","V2")
  
  mu <- exp(Fxtrain)
  ytrain <- sapply(mu, rand_tweedie, xi = rho, phi = phi)
  train_dat$Y <- ytrain
  
  cat("No of simulation:", i, "\n")
  
  # MGCV/TGAM
  t0 <- proc.time()
  m1<- gam(Y~s(V1)+s(V2),data = train_dat, family=Tweedie(p=rho, link = "log"))
  pred_f1 = predict.gam(m1, newdata = test_dat)
  timing1[i] <- (proc.time() - t0)[3]
  
  # TDboost
  t0 <- proc.time()
  ## cross-validation
  m2 <- TDboost(Y~V1+V2, data = train_dat,
                distribution = list(name="EDM", alpha=rho), # poisson, coxph, and quantile available
                shrinkage = 0.001,               # shrinkage or learning rate,
                n.trees = 30000,              # 0.001 to 0.1 usually work
                interaction.depth = 2,         # 1: additive model, 2: two-way interactions, etc.
                bag.fraction = 1,            # subsampling fraction, 0.5 is probably best
                n.minobsinnode = 20,		 # minimum total weight needed in each node
                keep.data = TRUE,              # keep a copy of the dataset with the object
                cv.folds = 5,
                verbose = FALSE)                # print out progress
  
  best.iter <- TDboost.perf(m2, method="cv", plot.it = FALSE)
  # predicting
  pred_f2 <- predict.TDboost(m2, test_dat, best.iter, type = "link")
  timing2[i] <- (proc.time() - t0)[3]
  
  # TGLM
  t0 <- proc.time()
  m3 <- glm(Y~V1+V2, data = train_dat,
            family = tweedie(link.power=0, var.power=rho))
  pred_f3 <- predict(m3, test_dat)
  timing3[i] <- (proc.time() - t0)[3]
  
  # ktweedie - RBF
  t0 <- proc.time()
  ## cross-validation
  sigma.range.rbf <- kernlab::sigest(xtrain)[c(1,3)]
  cv2d.rbf <- ktd_cv2d(x = xtrain,
                       y = ytrain,
                       kernfunc = rbfdot,
                       lambda = lambda.range.rbf,
                       sigma = sigma.range.rbf,
                       ncoefs = 15,
                       nfolds = 5,
                       ftol = 1e-4,
                       partol = 1e-4,
                       rho = rho,
                       verbose = FALSE)
  ## training
  m4 <- ktd_estimate(x = xtrain,
                     y = ytrain,
                     kern = rbfdot(cv2d.rbf$Best_sigma),
                     lam1 = cv2d.rbf$Best_lambda,
                     rho = rho,
                     partol = 1e-6,
                     ftol = 1e-6,
                     verbose = FALSE)
  ## predicting
  pred_m4 <- ktd_predict(model = m4, newdata = xtest, type = "link")
  pred_f4 <- as.vector(pred_m4$prediction)
  timing4[i] <- (proc.time() - t0)[3]
  
  # ktweedie - Laplace
  t0 <- proc.time()
  ## cross-validation
  cv2d.lap <- ktd_cv2d(x = xtrain,
                       y = ytrain,
                       kernfunc = laplacedot,
                       lambda = lambda.range.lap,
                       sigma = sigma.range.lap,
                       ncoefs = 15,
                       nfolds = 5,
                       ftol = 1e-4,
                       partol = 1e-4,
                       rho = rho,
                       verbose = FALSE
  )
  ## training
  m5 <- ktd_estimate(x = xtrain,
                     y = ytrain,
                     kern = laplacedot(cv2d.lap$Best_sigma),
                     lam1 = cv2d.lap$Best_lambda,
                     rho = rho,
                     ftol = 1e-6,
                     partol = 1e-6,
                     verbose = FALSE
  )
  ## testing
  pred_m5 <- ktd_predict(model = m5, newdata = xtest, type = "link")
  pred_f5 <- as.vector(pred_m5$prediction)
  timing5[i] <- (proc.time() - t0)[3]
  
  # Mean absolute difference (MAD) storage
  sim1[i] <- mean(abs(pred_f1 - truef))
  sim2[i] <- mean(abs(pred_f2 - truef))
  sim3[i] <- mean(abs(pred_f3 - truef))
  sim4[i] <- mean(abs(pred_f4 - truef))
  sim5[i] <- mean(abs(pred_f5 - truef))
  
  plot_pred_gam <- predict.gam(m1, newdata = plot_df)
  plot_pred_glm <- predict.glm(m3, newdata = plot_df)
  plot_pred_tdb <- predict.TDboost(m2,
                                   newdata = plot_df,
                                   best.iter,
                                   type = "link")
  plot_pred_rbf <- as.vector(ktd_predict(model = m4,
                                         newdata = plot_mat,
                                         type = "link")$prediction)
  plot_pred_lap <- as.vector(ktd_predict(model = m5,
                                         newdata = plot_mat,
                                         type = "link")$prediction)
  
  plot_data <- data.frame(plot_X1, plot_X2,
                          plot_true, plot_pred_gam,
                          plot_pred_glm, plot_pred_tdb,
                          plot_pred_rbf, plot_pred_lap)
  p1 <- wireframe(plot_true ~ plot_X1 + plot_X2,
                  scales=list(arrows=F,col=1),
                  data = plot_data,
                  shade=TRUE,
                  par.settings = list(axis.line = list(col = "transparent")),
                  xlab = "", ylab = "", zlab = "",
                  # xlab = "X1",
                  # ylab = "X2",
                  # zlab = "F(X1,X2)",
                  # cex.lab = 1.2,
                  main="TRUE")
  p2 <- wireframe(plot_pred_gam ~ plot_X1 + plot_X2,
                  scales=list(arrows=F,col=1),
                  data = plot_data,
                  shade=TRUE,
                  par.settings = list(axis.line = list(col = "transparent")),
                  xlab = "", ylab = "", zlab = "",
                  # xlab = "X1",
                  # ylab = "X2",
                  # zlab = "F(X1,X2)",
                  # cex.lab = 1.2,
                  main="MGCV")
  p3 <- wireframe(plot_pred_glm ~ plot_X1 + plot_X2,
                  scales=list(arrows=F,col=1),
                  data = plot_data,
                  shade=TRUE,
                  par.settings = list(axis.line = list(col = "transparent")),
                  xlab = "", ylab = "", zlab = "",
                  # xlab = "X1",
                  # ylab = "X2",
                  # zlab = "F(X1,X2)",
                  # cex.lab = 1.2,
                  main="TGLM")
  p4 <- wireframe(plot_pred_tdb ~ plot_X1 + plot_X2,
                  scales=list(arrows=F,col=1),
                  data = plot_data,
                  shade=TRUE,
                  par.settings = list(axis.line = list(col = "transparent")),
                  xlab = "", ylab = "", zlab = "",
                  # xlab = "X1",
                  # ylab = "X2",
                  # zlab = "F(X1,X2)",
                  # cex.lab = 1.2,
                  main="TDboost")
  p5 <- wireframe(plot_pred_rbf ~ plot_X1 + plot_X2,
                  scales=list(arrows=F,col=1),
                  data = plot_data,
                  shade=TRUE,
                  par.settings = list(axis.line = list(col = "transparent")),
                  xlab = "", ylab = "", zlab = "",
                  # xlab = "X1",
                  # ylab = "X2",
                  # zlab = "F(X1,X2)",
                  # cex.lab = 1.2,
                  main="ktweedie - RBF")
  p6 <- wireframe(plot_pred_lap ~ plot_X1 + plot_X2,
                  scales=list(arrows=F,col=1),
                  data = plot_data,
                  shade=TRUE,
                  par.settings = list(axis.line = list(col = "transparent")),
                  xlab = "", ylab = "", zlab = "",
                  # xlab = "X1",
                  # ylab = "X2",
                  # zlab = "F(X1,X2)",
                  # cex.lab = 1.2,
                  main="ktweedie - Laplace")
  grid.arrange(p1, p2, p3, p4, p5, p6,
               nrow = 2, ncol = 3)
}

# Sound alarm for completion
if (sound_alarm_when_finish) {
  beep(8)
}

# Compare results
MAD_mean <- c(mean(sim1), mean(sim2), mean(sim3), mean(sim4), mean(sim5))
MAD_se <- c(sd(sim1), sd(sim2), sd(sim3), sd(sim4), sd(sim5))/sqrt(n.sim - 1)

Time_mean <- c(mean(timing1),
               mean(timing2),
               mean(timing3),
               mean(timing4),
               mean(timing5))
Time_se <- c(sd(timing1),
             sd(timing2),
             sd(timing3),
             sd(timing4),
             sd(timing5))

models <- c("MGCV", "TDboost", "TGLM", "RBF", "Laplace")
names(MAD_mean) <- names(MAD_se) <- names(Time_mean) <- names(Time_se) <- models

# MAD results
kable(cbind(MAD_mean, MAD_se), format = "simple",
      caption = "Mean Absolute Difference")
# Timing results
kable(cbind(Time_mean, Time_se), format = "simple",
      caption = "Timing Results")









